import java.sql.Connection

import com.branegy.dbmaster.model.ModelDataSource
import com.branegy.dbmaster.model.RevEngineeringOptions
import com.branegy.dbmaster.model.ModelObject

import java.sql.Connection
import java.sql.ResultSet
import java.sql.SQLException
import java.sql.Statement
import java.util.HashMap
import java.util.List
import java.util.Map
import java.util.Set
import java.util.Map.Entry

import org.apache.commons.lang.StringUtils

import com.branegy.dbmaster.connection.AbstractJdbcDialect
import com.branegy.dbmaster.model.RevEngineeringOptions
import com.branegy.dbmaster.model.Table
import com.branegy.dbmaster.model.Trigger
import com.branegy.dbmaster.model.View
import com.branegy.dbmaster.model.ForeignKey.ReferentialAction
import com.branegy.dbmaster.model.Index
import com.branegy.dbmaster.model.Procedure
import com.branegy.dbmaster.model.Column
import com.branegy.dbmaster.model.Constraint
import com.branegy.dbmaster.model.DatabaseInfo
import com.branegy.dbmaster.model.DatabaseObject
import com.branegy.dbmaster.model.ForeignKey
import com.branegy.dbmaster.model.Function
import com.branegy.dbmaster.model.ForeignKey.ReferentialAction;
import com.branegy.dbmaster.model.ForeignKey.ColumnMapping;
import com.branegy.dbmaster.model.Index.IndexColumn;
import com.branegy.dbmaster.model.Index.IndexType;
import com.branegy.dbmaster.model.Parameter;
import com.branegy.dbmaster.model.Parameter.ParamType;
import com.branegy.dbmaster.util.NameMap
import com.branegy.inventory.model.Job
import com.branegy.persistence.custom.BaseCustomEntity
import com.branegy.service.core.exception.ApiException
import groovy.util.slurpersupport.GPathResult
import com.branegy.dbmaster.connection.*
import java.sql.PreparedStatement
import com.branegy.dbmaster.connection.AbstractJdbcDialect.PreparedSql

public abstract class SQLServerDialect extends AbstractJdbcDialect {
    // replace with filter! 
    final Set<String> excludeFunctionSet = ["dbo.fn_diagramobjects"] as Set;
    final Set<String> excludeTableSet = ["dbo.sysdiagrams"] as Set;
    
    public SQLServerDialect(JdbcConnector cp, Connection connection, String dialectName, String dialectVersion) 
            throws SQLException {
        super(cp, connection, dialectName, dialectVersion);
    }
    
    @Override
    protected ModelDataSource getModelDataSource(Connection connection, RevEngineeringOptions options) {
        ModelDataSource model = super.getModelDataSource(connection, options);
        if (isImport(options,"ExtendedProperty")) {
            populateExtendedProperties(connection, model, options);
        }
        return model;
    }
    
    protected void populateExtendedProperties(Connection connection, ModelDataSource model, RevEngineeringOptions options) {
        String sql =
        """
SELECT s.name as SchemaName, o.name as ObjectName, c.name as ColumnName, p.name as ParameterName, o.type,

   o.type_desc, ep.name as PropertyName, ep.value as PropertyValue 
-- SQL_VARIANT_PROPERTY(ep.value,'BaseType') AS 'Base Type', 
-- SQL_VARIANT_PROPERTY(ep.value,'Precision') AS 'Precision', 
-- SQL_VARIANT_PROPERTY(ep.value,'Scale') AS 'Scale'

 FROM sys.extended_properties ep 
 LEFT JOIN sys.all_objects o ON ep.major_id = o.object_id 
 LEFT JOIN sys.schemas s on o.schema_id = s.schema_id 
 LEFT JOIN sys.columns AS c ON ep.major_id = c.object_id AND ep.minor_id = c.column_id 
 LEFT JOIN sys.parameters as p on ep.major_id = p.object_id AND ep.minor_id = p.parameter_id
        """;
        java.sql.Statement st = null;
        ResultSet rs = null;
        try {
            st = connection.createStatement();
            rs = st.executeQuery(sql);
            while (rs.next()) {
                String schemaName = rs.getString("SchemaName");
                String objectName = rs.getString("ObjectName");
                String columnName = rs.getString("ColumnName");
                String parameterName = rs.getString("ParameterName");
                String type = rs.getString("type");
                String name = rs.getString("PropertyName");
                String value = rs.getString("PropertyValue");
              
                String fullName = schemaName!=null?schemaName+"."+objectName:objectName;
                DatabaseObject<?> target = null;
                switch (type){
                    case "U ":
                        target = model.getTable(fullName);
                        if (target!=null && columnName!=null) {
                            target = target.getColumn(columnName); 
                        }
                        break;
                    case "V ":  
                        target = model.getView(fullName);
                        if (target!=null && columnName!=null) {
                            target = target.getColumn(columnName);
                        }
                        break;
                    case "IF":
                        target = model.getFunction(fullName);
                        if (target!=null && parameterName!=null) {
                            target = target.getParameter(parameterName);
                        }
                        break;
                    case "P ":
                        target = model.getProcedure(fullName);
                        if (target!=null && parameterName!=null) {
                            target = target.getParameter(parameterName);
                        }
                        break;
                }
                if (target!=null) {
                    setCustomValue(target, "ep:"+name,value);
                }
            }
        } catch (ApiException e) {
            throw e
        } catch (Exception e) {
            throw new ConnectionApiException("Cannot load extended properties:"+e.getLocalizedMessage(), e)
        } finally {
            closeQuietly(rs);
            closeQuietly(st);
        }
    }

    /*
    public void saveExtendedProperties(DatabaseObject object) {
        if ("Table".equals(object.getDiscriminator())){
            for(Entry<String,Object> e:object.getCustomMap().entrySet()) {
                String name = e.getKey();
                if (name.startsWith("ep:")) {
                    String schema = object.getSchema();
                    String tableName = object.getSimpleName();
                    //result.put(name,e.getValue());
                }
            }
        }
    }*/
    
    @Override
    public List<Job> getJobs() {
        String sql =
        """
select 
	job.job_id, 
	job.name, 
	job.enabled, 
	job.description,
	c.name as category_name, 
	suser_sname(job.owner_sid) as owner,
	(select 
	job.date_created,
	job.date_modified,
	job.start_step_id,
	isnull((
		select 
			s.schedule_id,
			schedule_uid,
			originating_server_id,
			name,
			suser_sname(owner_sid) as owner,
			enabled,
			freq_type,
			freq_interval,
			freq_subday_type,
			freq_subday_interval,
			freq_relative_interval,
			freq_recurrence_factor,
			active_start_date,
			active_end_date,
			active_start_time,
			active_end_time,
			date_created,
			date_modified,
			version_number
		from msdb.dbo.sysjobschedules js 
			inner join msdb.dbo.sysschedules s on js.schedule_id = s.schedule_id 
		where js.job_id = job.job_id
		for xml path('Schedule'), type
	),'') as Schedules,
	isnull((
		select 
			1		              as Tag,
			NULL	              as Parent,
			step_id               as [Step!1!step_id!ELEMENT],
			step_name             as [Step!1!step_name!ELEMENT],
			subsystem             as [Step!1!subsystem!ELEMENT],
			command               as [Step!1!command!CDATA],
			cmdexec_success_code  as [Step!1!cmdexec_success_code!ELEMENT],
			flags                 as [Step!1!flags!ELEMENT],
			on_success_action     as [Step!1!on_success_action!ELEMENT],
			on_success_step_id    as [Step!1!on_success_step_id!ELEMENT],
			on_fail_action        as [Step!1!on_fail_action!ELEMENT],
			on_fail_step_id       as [Step!1!on_fail_step_id!ELEMENT],
			database_name         as [Step!1!database_name!ELEMENT],
			database_user_name    as [Step!1!database_user_name!ELEMENT],
			retry_attempts        as [Step!1!retry_attempts!ELEMENT], 
			retry_interval        as [Step!1!retry_interval!ELEMENT],  
			output_file_name      as [Step!1!output_file_name!ELEMENT],  
			proxy_id              as [Step!1!proxy_id!ELEMENT]  
		from 
			msdb.dbo.sysjobsteps st
		where st.job_id = job.job_id
		for xml explicit, type
	),'')	 as Steps
	   for xml path('JobDefinition'), type
	) as JobDefinition
from msdb.dbo.sysjobs job 
	left join msdb.dbo.syscategories c on job.category_id=c.category_id
""";

        try {
            Connection connection = getConnection()
            List<Job> result = new ArrayList<Job>()
            java.sql.Statement st = null;
            ResultSet rs = null
            try{
                st = connection.createStatement()
                rs = st.executeQuery(sql)
                while (rs.next()) {
                    Job job = new Job()
                    job.setJobType("SqlServerJob")
                    job.setJobName(rs.getString("name"))

                    job.setCustomData("JobId",         rs.getString("job_id"))
                    job.setCustomData("Enabled",       rs.getBoolean("enabled"))
                    job.setCustomData("Source",        "SqlServer")
                    job.setCustomData("Description",   rs.getString("description"))
                    job.setCustomData("Category",      rs.getString("category_name"))
                    job.setCustomData("Owner",         rs.getString("owner"))
                    job.setCustomData("JobDefinition", rs.getString("JobDefinition"))                    
                    result.add(job)
                }
            } finally{
                closeQuietly(rs);
                closeQuietly(st);
            }
            return result
        } catch (ApiException e) {
            throw e
        } catch (Exception e) {
            throw new ConnectionApiException("Cannot load jobs:"+e.getLocalizedMessage(), e)
        }
    }

    
    
    protected NameMap<Table> getTables(Connection conn,
                                       RevEngineeringOptions options,
                                       String sqlWithWhere,
                                       String sqlOrderSuffix,
                                       String schemaColumn,
                                       String objectColumn,
                                       Map.Entry<?,Class<?>>... extraCustomFields 
        ) throws Exception {
            
        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, Table.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;  
        }
    
        NameMap<Table> tables = new NameMap<Table>(200);
        PreparedStatement statement = null;
        ResultSet tablesRs = null;
        try{
            statement = conn.prepareStatement(sql);
            filter.populatePreparedStatement(statement);
            tablesRs = statement.executeQuery();
            while (tablesRs.next()) {
                String schemaName = tablesRs.getString("schemaName");
                String tableName = tablesRs.getString("tableName");
                
                if (excludeTableSet.contains(schemaName+"."+tableName)){
                    continue;
                }
        
                Table tableInfo = new Table();
                tableInfo.setName(schemaName+"."+tableName);
                
                for (Map.Entry<?,Class<?>> extra:extraCustomFields) {
                    tableInfo.setCustomData(extra.getKey(), tablesRs.getObject(extra.getKey(),extra.getValue()));
                }
                tables.put(schemaName + "." + tableName, tableInfo);
            }
        } finally {
            closeQuietly(tablesRs);
            closeQuietly(statement);
        }
        if (isImport(options,Column.class)){
            populateTableColumns(conn, options, tables);
        }
        if (isImport(options,Constraint.class)){
            populateConstraints(conn, options, tables);
        }
        if (isImport(options,ForeignKey.class)){
            populateForeignKeys(conn, options, tables);
        }
        if (isImport(options,Index.class)){
            populateIndexes(conn, options, tables);
            populateIndexColumns(conn, options, tables);
        }
        if (isImport(options,Trigger.class)){
            populateTriggers(conn, options, tables);
        }
        return tables;
    }
    
    private static void setCustomValue(BaseCustomEntity o, String fieldName, Object value) {
        if ((value instanceof String) && StringUtils.isBlank((String)value)) {
            value = null
        }
        o.setCustomData(fieldName, value)
    }
    
    private boolean isUnicode(String type) {
        return type.equals("nvarchar") || type.equals("nchar");
    }
    
    protected void populateTableColumns(Connection conn, 
                                        RevEngineeringOptions options, 
                                        NameMap<Table> tables,
                                        String sqlWithWhere,
                                        String sqlOrderSuffix,
                                        String schemaColumn,
                                        String objectColumn
                                        ) {
        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, Column.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;  
        }
        
        PreparedStatement statement = null;
        ResultSet columnsRS = null;
        try{
            statement = conn.prepareStatement(sql)
            filter.populatePreparedStatement(statement);
            columnsRS = statement.executeQuery();
            while (columnsRS.next()) {
                String schema = columnsRS.getString("schema_name");
                String table = columnsRS.getString("table_name");
                Table tableInfo = tables.get(schema + "." + table);
                if (tableInfo == null) {
                    continue;
                    // throw new RuntimeException("Table ${schema}.${table} was not found in the model");
                }
                  
                Column columnInfo = new Column();
                columnInfo.setName(columnsRS.getString("column_name"));
          
                String type = columnsRS.getString("type_name");
                columnInfo.setType(type);
                     
                if (typesWithSize.contains(type.toUpperCase())) {
                    int size = columnsRS.getInt("max_length");
                    if (size > 0 && isUnicode(type)) {// size that equals -1  means unlimited (max) length
                        size /= 2;
                    }
                    columnInfo.setSize(size);
                }
          
                if (typesWithPrecesion.contains(type.toUpperCase())) {
                    columnInfo.setPrecesion(columnsRS.getInt("precision"));
                }
          
                if (typesWithScale.contains(type.toUpperCase())) {
                    columnInfo.setPrecesion(columnsRS.getInt("precision"));
                    columnInfo.setScale(columnsRS.getInt("scale"));
                }
          
                String nullable = columnsRS.getString("nullable");
                columnInfo.setNullable(nullable != null && nullable.equals("1"));
          
                String string = columnsRS.getString("defaultValue");
                columnInfo.setDefaultValue(string);
          
                tableInfo.addColumn(columnInfo);
                // TODO columnInfo.setExtraDefinition(extra);
          
                setCustomValue(columnInfo, "is_identity", columnsRS.getInt("is_identity"))
          
                setCustomValue(columnInfo, "is_computed", columnsRS.getInt("is_computed"))
                setCustomValue(columnInfo, "computedExpression", columnsRS.getString("computedExpression"))
                setCustomValue(columnInfo, "is_persisted", columnsRS.getObject("is_persisted"))
                setCustomValue(columnInfo, "uses_database_collation", columnsRS.getObject("uses_database_collation"))
              
                setCustomValue(columnInfo, "is_rowguidcol", columnsRS.getInt("is_rowguidcol"))
                setCustomValue(columnInfo, "is_user_defined", columnsRS.getInt("is_user_defined"))
                setCustomValue(columnInfo, "user_type", columnsRS.getString("user_type"))
                setCustomValue(columnInfo, "collation_name", columnsRS.getString("collation_name"))
          }
        } finally {
            closeQuietly(columnsRS);
            closeQuietly(statement);
        }
    } 
    
    /**
     * Populates table and column level constraints
     */
    protected void populateConstraints(Connection conn, RevEngineeringOptions options, NameMap<Table> tables) throws Exception {
        String sql =
      """select s.name as schemaName,o.name as tableName,c.name as columnName,
                   cc.name as constraintName, cc.is_disabled, cc.definition 
         from sys.check_constraints as cc 
         inner join sys.objects as o on o.object_id = cc.parent_object_id 
         inner join sys.schemas as s on s.schema_id = o.schema_id 
         left join sys.columns as c on o.object_id = c.object_id and cc.parent_column_id = c.column_id
         where""";

        this.populateConstraints(conn,options,tables, sql, null, "s.name","o.name");
    }
    
    /**
     * Populates table and column level constraints
     */
    protected void populateConstraints( Connection conn, 
                                        RevEngineeringOptions options, 
                                        NameMap<Table> tables,
                                        String sqlWithWhere,
                                        String sqlOrderSuffix,
                                        String schemaColumn,
                                        String objectColumn) throws Exception {
         
        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, Constraint.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;
        }
            
        PreparedStatement statement = null;
        ResultSet constraintsRS = null;
        try{
            statement = conn.prepareStatement(sql);
            filter.populatePreparedStatement(statement);
            constraintsRS = statement.executeQuery();
            while (constraintsRS.next()) {
                String tableName =
                        constraintsRS.getString("schemaName") + "." + constraintsRS.getString("tableName");
    
                Table table = tables.get(tableName);
                if (table == null) {
                    // table is not a part of reverse engineering
                    continue;
                }
                Constraint constraint = new Constraint();
                constraint.setName(constraintsRS.getString("constraintName"));
                constraint.setDisabled(constraintsRS.getBoolean("is_disabled"));
                constraint.setDefinition(constraintsRS.getString("definition"));
                constraint.setColumnName(constraintsRS.getString("columnName"));
    
                table.addConstraint(constraint);
            }
        } finally {
            closeQuietly(constraintsRS);
            closeQuietly(statement);
        }
    }
    
    protected ReferentialAction getReferentialAction(int action) {
        switch (action) {
        case 0:
            return ReferentialAction.NoAction;
        case 1:
            return ReferentialAction.Cascade;
        case 2:
            return ReferentialAction.SetNull;
        case 3:
            return ReferentialAction.SetDefault;
        default:
            // TODO Add logging
            return ReferentialAction.UnknownAction;
        }
    }
    
    /**
     * Populates foreign keys (including columns)
     */
    protected void populateForeignKeys(Connection conn, RevEngineeringOptions options, NameMap<Table> tables) throws Exception {
        String sql =
        """select s.name as schemaName,
                          o.name as tableName,
                          c.name as columnName, 
                          fk.name keyName, 
                          fk.is_disabled, 
                          fk.delete_referential_action, 
                          fk.update_referential_action, 
                          fkc.constraint_column_id columnNumber, 
                          ref_s.name as refSchema, 
                          ref_o.name as refTable,
                          ref_c.name as refColumn 
           from sys.foreign_keys fk 
         inner join sys.foreign_key_columns fkc on fk.object_id = fkc.constraint_object_id 
         inner join sys.objects o on fkc.parent_object_id = o.object_id 
         inner join sys.columns c on o.object_id = c.object_id and fkc.parent_column_id = c.column_id 
         inner join sys.schemas s on o.schema_id = s.schema_id 
         inner join sys.objects ref_o on fkc.referenced_object_id = ref_o.object_id 
         inner join sys.columns ref_c on ref_o.object_id = ref_c.object_id 
                                and fkc.referenced_column_id = ref_c.column_id 
         inner join sys.schemas ref_s on ref_o.schema_id = ref_s.schema_id
         where""";
         
         this.populateForeignKeys(conn, options, tables, sql, "order by schemaName,tableName,keyName,columnNumber",
              "s.name","c.name");
    }
    
    protected void populateForeignKeys(Connection conn, RevEngineeringOptions options, NameMap<Table> tables,
        String sqlWithWhere,
        String sqlOrderSuffix,
        String schemaColumn,
        String objectColumn) throws Exception {

        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, ForeignKey.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;
        }
        
        PreparedStatement statement = null;
        ResultSet keysRS = null;
        try{
            statement = conn.prepareStatement(sql);
            filter.populatePreparedStatement(statement);
            keysRS = statement.executeQuery();
            ForeignKey key = null;
            while (keysRS.next()) {
                String tableName = keysRS.getString("schemaName") + "." + keysRS.getString("tableName");
    
                Table table = tables.get(tableName);
                if (table == null) {
                    // table is not a part of reverse engineering
                    continue;
                }
                String keyName = keysRS.getString("keyName");
                if (key == null || !key.getName().equals(keyName)) {
                    key = new ForeignKey();
                    key.setName(keyName);
                    key.setDisabled(keysRS.getBoolean("is_disabled"));
                    key.setDeleteAction(getReferentialAction(keysRS.getInt("delete_referential_action")));
                    key.setUpdateAction(getReferentialAction(keysRS.getInt("update_referential_action")));
    
                    key.setTargetTable(keysRS.getString("refSchema") + "." + keysRS.getString("refTable"));
    
                    table.addForeignKey(key);
                }
                String sourceColumn = keysRS.getString("columnName");
                String targetColumn = keysRS.getString("refColumn");
                ColumnMapping mapping = new ColumnMapping(sourceColumn, targetColumn);
                key.addColumnMapping(mapping);
    
            }
        } finally {
            closeQuietly(keysRS);
            closeQuietly(statement);
        }
    }
 
    protected void populateIndexes(Connection conn, RevEngineeringOptions options, NameMap<Table> tables) {
        String objType = 'U';
        
        String sql =
        """SELECT s.name as schemaName,
                        o.name as objectName,
                        o.type as objectType, 
                        i.name as indexName,
                        i.type indexType,
                        i.is_unique,
                        i.ignore_dup_key, 
                        i.is_primary_key, 
                        i.is_unique_constraint,
                        i.is_disabled,
                        i.fill_factor
                from  sys.objects o   
                inner join sys.indexes i on i.object_id = o.object_id  
                inner join sys.schemas s on o.schema_id = s.schema_id  
                where i.name IS NOT NULL 
                  AND i.is_hypothetical=0 and o.type='${objType}' AND""";
        this.populateIndexes(conn, options, tables, sql, null,"s.name","i.name");     
    }
    
     /**
     * @objType can be U: USER TABLE, V: View, TF: Table valued functions
     * Requires View server state permission
     */
    protected void populateIndexes(Connection conn, RevEngineeringOptions options, NameMap<Table> tables, 
        String sqlWithWhere,
        String sqlOrderSuffix,
        String schemaColumn,
        String objectColumn) {
        
        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, Index.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;
        }
        
        PreparedStatement statement = null;
        ResultSet indexRS = null;
        try{
            statement = conn.prepareStatement(sql)
            filter.populatePreparedStatement(statement);
            indexRS = statement.executeQuery()
            Index index = null;
            while (indexRS.next()) {
                String tableName = indexRS.getString("schemaName") + "." + indexRS.getString("objectName");
                Table object = tables.get(tableName);
                if (object == null) {
                    // object is not a part of reverse engineering
                    continue;
                }
            
                String indexName = indexRS.getString("indexName");
                index = new Index();
                index.setName(indexName);
                index.setUnique(indexRS.getInt("is_unique_constraint") == 1);
                index.setPrimaryKey(indexRS.getInt("is_primary_key") == 1);
                index.setFillFactor(indexRS.getInt("fill_factor"));
                index.setDisabled(indexRS.getInt("is_disabled") == 1);
                index.setIgnoreDuplicates(indexRS.getInt("ignore_dup_key") == 1);
                index.setType(getIndexType(indexRS.getInt("indexType")));
            
                object.addIndex(index);
            }
        } finally {
            closeQuietly(indexRS);
            closeQuietly(statement);
        }
    }
    
    protected void populateIndexColumns(Connection conn, RevEngineeringOptions options, NameMap<Table> tables) {
        String objType = 'U';
        String sql =
          """ select s.name as schemaName,
                     o.name objectName, 
                     i.name indexName,
                     c.name as columnName,  
                     ic.key_ordinal,
                     ic.is_descending_key,
                     ic.is_included_column  
              from sys.indexes as i  
              inner join sys.index_columns as ic on i.object_id = ic.object_id and i.index_id = ic.index_id  
              inner join sys.objects as o on i.object_id = o.object_id  
              inner join sys.schemas s on o.schema_id = s.schema_id  
              inner join sys.columns c on o.object_id = c.object_id and ic.column_id = c.column_id  
              where o.type='${objType}' and i.name is not null
              order by schemaName,objectName,indexName,key_ordinal """;
        
        Statement statement = null;
        ResultSet indexRS = null;
        try{
            statement = conn.createStatement();
            indexRS = statement.executeQuery(sql)
            Index index = null;
            while (indexRS.next()) {
                String schemaName = indexRS.getString("schemaName");
                String tableName = schemaName + "." + indexRS.getString("objectName");
    
                Table object = tables.get(tableName);
                if (object == null) {
                    // object is not a part of reverse engineering
                    continue;
                }
    
                String indexName = indexRS.getString("indexName");
                index = object.getIndex(indexName);
    
                IndexColumn indexColumn = new IndexColumn();
                indexColumn.setAsc(indexRS.getInt("is_descending_key") == 0);
                indexColumn.setColumnName(indexRS.getString("columnName"));
                indexColumn.setIncluded(indexRS.getInt("is_included_column") == 1);
    
                index.addIndexColumn(indexColumn);
            }
        } finally {
            closeQuietly(indexRS);
            closeQuietly(statement);
        }
    }
    
    protected IndexType getIndexType(int sqlIndexType) {
        switch (sqlIndexType) {
        case 0:
            return IndexType.Heap;
        case 1:
            return IndexType.Clustered;
        case 2:
            return IndexType.Nonclustered;
        case 3:
            return IndexType.XML;
        case 4:
            return IndexType.Spatial;
        default:
            return IndexType.UnknownIndexType;
        }
    }
    
    /**
     * populate database triggers
     */
    protected void populateTriggers(Connection conn, RevEngineeringOptions options, NameMap<Table> objects)
            throws Exception {
        String sql =
              """   select
                    tr.object_id as trigger_id,
                    tr.parent_id as parent_id,
                    o.name as trigger_name ,
                    user_name(o.uid) as trigger_owner,
                    s.name as table_schema,
                    object_name(parent_obj) as table_name,
                    objectproperty( id, 'ExecIsUpdateTrigger') as isupdate,
                    objectproperty( id, 'ExecIsDeleteTrigger') as isdelete,
                    objectproperty( id, 'ExecIsInsertTrigger') as isinsert,
                    objectproperty( id, 'ExecIsAfterTrigger') as isafter,
                    objectproperty( id, 'ExecIsInsteadOfTrigger') as isinsteadof,
                    objectproperty(id, 'ExecIsTriggerDisabled') as [disabled],
                    object_definition( tr.object_id ) as 'source'
                    from sysobjects o
                    inner join sysusers on o.uid = sysusers.uid
                    inner join sys.tables t on o.parent_obj = t.object_id
                    inner join sys.schemas s on t.schema_id = s.schema_id
                    inner join sys.triggers tr on o.id = tr.object_id
                    where
                    tr.type = 'tr'
              """;

        Statement statement = null;
        ResultSet triggerRS = null;
        try{
            statement = conn.createStatement();
            triggerRS = statement.executeQuery(sql);
            Trigger trigger = null;
            while (triggerRS.next()) {
                String tableName = triggerRS.getString("table_schema") + "." + triggerRS.getString("table_name");
                Table object = objects.get(tableName);
                if (object == null) {
                    // object is not a part of reverse engineering
                    continue;
                }
    
                trigger = new Trigger();
                trigger.setTriggerId(triggerRS.getInt("trigger_id"));
                trigger.setParentId(triggerRS.getInt("parent_id"));
                trigger.setName(triggerRS.getString("trigger_name"));
                trigger.setTriggerOwner(triggerRS.getString("trigger_owner"));
                trigger.setUpdate(triggerRS.getInt("isupdate") == 1);
                trigger.setDelete(triggerRS.getInt("isdelete") == 1);
                trigger.setInsert(triggerRS.getInt("isinsert") == 1);
                trigger.setAfter(triggerRS.getInt("isafter") == 1);
                trigger.setInsteadOf(triggerRS.getInt("isinsteadof") == 1);
                trigger.setDisabled(triggerRS.getInt("disabled") == 1);
                trigger.setSource(triggerRS.getString("source"));
                
                object.addTrigger(trigger);
            }
        } finally {
            closeQuietly(triggerRS);
            closeQuietly(statement);
        }
    }
    
    @Override
    protected NameMap<Function> getFunctions(Connection conn, RevEngineeringOptions options) {
        String sql =
                 """select
                        s.name as schemaName,  
                        obj.name as functionName,
                        obj.create_date, 
                        obj.modify_date,
                        case obj.type 
                            when N'FN' then 'Scalar'
                            when N'IF' then 'Inline table-valued'
                            when N'TF' then 'Table-valued'
                            when N'FS' then 'Assembly scalar'
                            when N'FT' then 'Assembly table-valued'
                        end 
                            as functionType,
                        m.definition 
                    from sys.objects as obj
                    inner join sys.schemas as s on obj.schema_id=s.schema_id 
                    left  join sys.sql_modules as m on obj.object_id=m.object_id
                    where obj.type in (N'FN', N'IF', N'TF', N'FS', N'FT')""";

        PreparedSql filter = getSqlFilter("s.name", "obj.name", Function.class, options);
        sql += " and " + filter.getSql() + " order by s.name, obj.name";

        NameMap<Function> functions = null;
        PreparedStatement statement = null;
        ResultSet functionsRs = null;
        try{
            statement = conn.prepareStatement(sql)
            filter.populatePreparedStatement(statement);
            functionsRs = statement.executeQuery();
            functions = new NameMap<Function>(100);
            while (functionsRs.next()) {
                String schemaName = functionsRs.getString("schemaName");
                String functionName = functionsRs.getString("functionName");
                
                if (excludeFunctionSet.contains(schemaName+"."+functionName)){
                    continue;
                }
                
                Function function = functions.get(schemaName+"."+functionName);
                String definition = functionsRs.getString("definition");
                if (function != null) {
                    function.setSource(function.getSource() + definition);
                } else {
                    function = new Function();
                    function.setSource(definition);
                    function.setName(schemaName+"."+functionName);
                    function.setType(functionsRs.getString("functionType"));
                    //String comments = functionsRs.getString("procedureDescription");
                    //procedure.setCustomData(Column.DESCRIPTION, comments);
                    functions.put(schemaName+"."+functionName, function);
                }
            }
        } finally {
            closeQuietly(functionsRs);
            closeQuietly(statement);
        }
        populateParameters(conn, options, functions, "N'FN', N'IF', N'TF', N'FS', N'FT'");
        return functions;
    }
    
    protected abstract void populateParameters(Connection conn, RevEngineeringOptions options,
        NameMap dbObjects, String typeFilter) throws SQLException;
    
    
    protected void populateParameters(Connection conn, RevEngineeringOptions options,
        NameMap dbObjects, String typeFilter,
        String sqlWithWhere,
        String sqlOrderSuffix,
        String schemaColumn,
        String objectColumn) throws SQLException {

        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, Parameter.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;
        }
        
        PreparedStatement statement = null;
        ResultSet paramRS = null;
        try{
            statement = conn.prepareStatement(sql);
            filter.populatePreparedStatement(statement);
            paramRS = statement.executeQuery();
            while(paramRS.next())
            {
                String schema = paramRS.getString("schema_name");
                String objectName = paramRS.getString("object_name");
                def obj = dbObjects.get(schema + "." + objectName);
                if (obj == null) {
                    continue;
                }
                
                Parameter paramInfo = new Parameter();
                String parameterName = paramRS.getString("column_name")
                if (parameterName.length()==0) {
                    parameterName = "RESULT"
                }
                paramInfo.setName(parameterName)
                String type = paramRS.getString("type_name")
                paramInfo.setType(type)
                if (typesWithSize.contains(type.toUpperCase())) {
                    int size = paramRS.getInt("max_length")
                    if (size>0 && isUnicode(type)) {
                        size /= 2;
                    }
                    paramInfo.setSize(size);
                }
                
                if (typesWithScale.contains(type.toUpperCase())) {
                    paramInfo.setPrecesion(paramRS.getInt("precision"));
                    paramInfo.setScale(paramRS.getInt("scale"));
                }
                
                paramInfo.setNullable(true);
                
                String string = paramRS.getString("defaultValue");
                paramInfo.setDefaultValue(string);
                
                obj.addParameter(paramInfo);
                String extra = "";
                //            collation_name
                if (paramRS.getInt("is_xml_document") == 1) {
                    extra += " XML Document";
                }
                if (paramRS.getInt("is_readonly") == 1) {
                    extra += " Readonly";
                }
                if (paramRS.getInt("is_output") == 1) {
                    paramInfo.setParamType(ParamType.OUT);
                } else {
                    paramInfo.setParamType(ParamType.IN);
                }
                paramInfo.setExtraDefinition(extra);
                //paramInfo.setCustomData(Column.DESCRIPTION, paramRS.getString("comment"));
            }
        } finally {
            closeQuietly(paramRS);
            closeQuietly(statement);
        }
    }
    
    @Override
    protected NameMap<Procedure> getProcedures(Connection conn, RevEngineeringOptions options) throws Exception {
        String sql =
                """select
                    s.name as schemaName,  
                    pr.name as procedureName,
                    pr.create_date, 
                    pr.modify_date,
                    m.definition as procedureText 
                    from sys.procedures as pr
                    inner join sys.schemas as s on pr.schema_id=s.schema_id 
                    left  join sys.sql_modules as m on pr.object_id=m.object_id
                    where""";
        return this.getProcedures(conn, options, sql, "order by s.name, pr.name", "s.name", "pr.name",
             "N'P',N'PC',N'X',N'RF'");
    }
    
    protected NameMap<Procedure> getProcedures(Connection conn, RevEngineeringOptions options,
        String sqlWithWhere,
        String sqlOrderSuffix,
        String schemaColumn,
        String objectColumn,
        String parametersTypeFilter) throws Exception {
        
        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, Procedure.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;
        }
       
        NameMap<Procedure> procedures = new NameMap<Procedure>(100);
        PreparedStatement statement = null;
        ResultSet proceduresRs = null;
        try{
            statement = conn.prepareStatement(sql);
            filter.populatePreparedStatement(statement);
            proceduresRs = statement.executeQuery();
            while (proceduresRs.next()) {
                String schemaName = proceduresRs.getString("schemaName");
                String procedureName = proceduresRs.getString("procedureName");
                
                //if (excludeProcedureSet.contains(schemaName+"."+procedureName)){
                //    continue;
                //}
                
                Procedure procedure = procedures.get(schemaName+"."+procedureName);
                String procedureText = proceduresRs.getString("procedureText");
                if (procedure != null) {
                    procedure.setSource(procedure.getSource() + procedureText);
                } else {
                    procedure = new Procedure();
                    procedure.setSource(procedureText);
                    procedure.setName(schemaName+"."+procedureName);
                    
                    String comments = proceduresRs.getString("procedureDescription");
                    procedure.setCustomData(Column.DESCRIPTION, comments);
                    procedures.put(schemaName+"."+procedureName, procedure);
                }
            }
        } finally {
            closeQuietly(proceduresRs);
            closeQuietly(statement);
        }
        populateParameters(conn, options, procedures,parametersTypeFilter);
        return procedures;
    }
    
    protected NameMap<View> getViews(Connection conn, RevEngineeringOptions options,
        String sqlWithWhere,
        String sqlOrderSuffix,
        String schemaColumn,
        String objectColumn) throws Exception {
        
        PreparedSql filter = getSqlFilter(schemaColumn, objectColumn, View.class, options);
        String sql = sqlWithWhere + " "+filter.getSql();
        if (sqlOrderSuffix!=null) {
            sql += " " + sqlOrderSuffix;
        }

        NameMap<View> views = null;
        PreparedStatement statement = null;
        ResultSet viewsRs = null;
        try{
            statement = conn.prepareStatement(sql)
            filter.populatePreparedStatement(statement);
            viewsRs = statement.executeQuery();
            views = new NameMap<View>(100);
            while (viewsRs.next()) {
                String schemaName = viewsRs.getString("schemaName")
                String viewName = viewsRs.getString("viewName")
                View view = views.get(viewName)
                String viewText = rsToString(viewsRs.getObject("viewText"))
                if (view != null) {
                    view.setSource(view.getSource() + viewText);
                } else {
                    view = new View()
                    view.setSource(viewText)
                    view.setName(schemaName+"."+viewName)
                    String comments = viewsRs.getString("viewDescription")
                    view.setCustomData(Column.DESCRIPTION, comments)
                    views.put(schemaName+"."+viewName, view)
                }
            }
        } finally {
            closeQuietly(viewsRs);
            closeQuietly(statement);
        }
        populateViewColumns(conn, options, views);
        return views;
    }
    
    protected void populateViewColumns(Connection conn, RevEngineeringOptions options,
        NameMap<View> views) throws SQLException {
        /*def metaData = conn.getMetaData()
        
        for (View view : views.values()) {
            ResultSet columnsRS = null
            try {
                String viewName = view.getName();
                columnsRS = metaData.getColumns(options.database, view.getSchema(), view.getSimpleName(), null);
                while (columnsRS.next()) {
    
                    Column column = new Column();
                    column.setName(columnsRS.getString("COLUMN_NAME"));
                    // int type = columns.getInt("DATA_TYPE"); // int => SQL type from
                    // java.sql.Types
                    String type = columnsRS.getString("TYPE_NAME");
                    column.setType(type);
                    if (typesWithSize.contains(type.toUpperCase())) {
                        column.setSize(columnsRS.getInt("COLUMN_SIZE"));
                    }
    
                    if (typesWithScale.contains(type.toUpperCase())) {
                        column.setPrecesion(column.getSize());
                        column.setScale(columnsRS.getInt("DECIMAL_DIGITS"));
                    }
    
                    String nullable = columnsRS.getString("NULLABLE");
                    column.setNullable(nullable != null && nullable.equals("1"));
    
                    String string = columnsRS.getString("COLUMN_DEF");
                    column.setDefaultValue(string);
    
                    // boolean unsigned = typeName.endsWith(" unsigned");
                    view.addColumn(column);
                }
            } catch (SQLException e) {
                logger.error("",e);
            } finally {
                columnsRS.close();
            }
        }
        setViewColumnComments(conn, options.database, views);
        views.values().each { it.updateCollectionOrder() }*/
            
            String sql =
            """select   sch.name as schema_name,
                            vew.name as view_name,
                            col.name as column_name,  
                            systyp.name as type_name, 
                            typ.is_user_defined, 
                            typ.name as user_type, 
                            col.max_length, 
                            col.precision, 
                            col.scale, 
                            col.collation_name, 
                            col.is_nullable as nullable,
                            col.is_identity,
                            col.is_computed,
                            cc.definition as computedExpression,
                            cc.is_persisted,
                            cc.uses_database_collation,
                            col.is_rowguidcol,
                            def.definition as defaultValue,
                            null as procedureDescription
                     from sys.columns as col  
                     inner join sys.views as vew on col.object_id=vew.object_id  
                     inner join sys.schemas as sch on vew.schema_id = sch.schema_id  
                     inner join sys.types as typ on col.user_type_id= typ.user_type_id  
                     inner join sys.types as systyp on col.system_type_id  = systyp.system_type_id  
                                and systyp.is_user_defined=0 and systyp.user_type_id = systyp.system_type_id  
                     left join sys.default_constraints as def on col.default_object_id = def.object_id 
                     left join sys.computed_columns as cc 
                            on  vew.object_id=cc.object_id and col.column_id = cc.column_id""";
    
          PreparedSql filter = getSqlFilter("sch.name", "vew.name", Column.class, options)
          sql +=" where " + filter.getSql()+" order by schema_name, view_name, col.column_id";
          
          PreparedStatement statement = null;
          ResultSet columnsRS = null;
          try{
              statement = conn.prepareStatement(sql)
              filter.populatePreparedStatement(statement);
              columnsRS = statement.executeQuery()
              while (columnsRS.next()) {
                  
                  String schema = columnsRS.getString("schema_name");
                  String view = columnsRS.getString("view_name");
                  View viewInfo = views.get(schema + "." + view);
                  if (viewInfo == null) {
                      continue;
                      // throw new RuntimeException("Table ${schema}.${table} was not found in the model");
                  }
                  Column columnInfo = new Column();
                  columnInfo.setName(columnsRS.getString("column_name"));
        
                  String type = columnsRS.getString("type_name");
                  columnInfo.setType(type);
                             
                  if (typesWithSize.contains(type.toUpperCase())) {
                      int size = columnsRS.getInt("max_length");
                      if (size > 0 && isUnicode(type)) {// size that equals -1  means unlimited (max) length
                          size /= 2;
                      }
                      columnInfo.setSize(size);
                  }
        
                  if (typesWithPrecesion.contains(type.toUpperCase())) {
                      columnInfo.setPrecesion(columnsRS.getInt("precision"));
                  }
        
                  if (typesWithScale.contains(type.toUpperCase())) {
                      columnInfo.setPrecesion(columnsRS.getInt("precision"));
                      columnInfo.setScale(columnsRS.getInt("scale"));
                  }
        
                  String nullable = columnsRS.getString("nullable");
                  columnInfo.setNullable(nullable != null && nullable.equals("1"));
        
                  String string = columnsRS.getString("defaultValue");
                  columnInfo.setDefaultValue(string);
        
                  viewInfo.addColumn(columnInfo);
                  // TODO columnInfo.setExtraDefinition(extra);
        
                  setCustomValue(columnInfo, "is_identity", columnsRS.getInt("is_identity"))
        
                  setCustomValue(columnInfo, "is_computed", columnsRS.getInt("is_computed"))
                  setCustomValue(columnInfo, "computedExpression", columnsRS.getString("computedExpression"))
                  setCustomValue(columnInfo, "is_persisted", columnsRS.getObject("is_persisted"))
                  setCustomValue(columnInfo, "uses_database_collation", columnsRS.getObject("uses_database_collation"))
        
                  setCustomValue(columnInfo, "is_rowguidcol", columnsRS.getInt("is_rowguidcol"))
                  setCustomValue(columnInfo, "is_user_defined", columnsRS.getInt("is_user_defined"))
                  setCustomValue(columnInfo, "user_type", columnsRS.getString("user_type"))
                  setCustomValue(columnInfo, "collation_name", columnsRS.getString("collation_name"))
              }
          } finally {
              closeQuietly(columnsRS);
              closeQuietly(statement);
          }
    }
    
    public List<DatabaseInfo> getDatabases() {
        return getDatabases(getConnection(), 
"""
select
    d.database_id,
    d.name,
    [Create Date]=create_date,
    [Compatibility Level]=compatibility_level,
    [Readonly]=case when is_read_only=0 then 'No' else 'Yes' end,
    [State]=state_desc,
    [Recovery Model]=recovery_model_desc,
    (
        select
            (
            select
                DATABASEPROPERTYEX(d.name, 'Collation') as Collation,
                DATABASEPROPERTYEX(d.name, 'ComparisonStyle') as ComparisonStyle,
                DATABASEPROPERTYEX(d.name, 'Edition') as Edition,
                DATABASEPROPERTYEX(d.name, 'IsAnsiNullDefault') as IsAnsiNullDefault,
                DATABASEPROPERTYEX(d.name, 'IsAnsiNullsEnabled') as IsAnsiNullsEnabled,
                DATABASEPROPERTYEX(d.name, 'IsAnsiPaddingEnabled') as IsAnsiPaddingEnabled,
                DATABASEPROPERTYEX(d.name, 'IsAnsiWarningsEnabled') as IsAnsiWarningsEnabled,
                DATABASEPROPERTYEX(d.name, 'IsArithmeticAbortEnabled') as IsArithmeticAbortEnabled,
                DATABASEPROPERTYEX(d.name, 'IsAutoClose') as IsAutoClose,
                DATABASEPROPERTYEX(d.name, 'IsAutoCreateStatistics') as IsAutoCreateStatistics,
                DATABASEPROPERTYEX(d.name, 'IsAutoCreateStatisticsIncremental') as IsAutoCreateStatisticsIncremental,
                DATABASEPROPERTYEX(d.name, 'IsAutoShrink') as IsAutoShrink,
                DATABASEPROPERTYEX(d.name, 'IsAutoUpdateStatistics') as IsAutoUpdateStatistics,
                DATABASEPROPERTYEX(d.name, 'IsClone') as IsClone,
                DATABASEPROPERTYEX(d.name, 'IsCloseCursorsOnCommitEnabled') as IsCloseCursorsOnCommitEnabled,
                DATABASEPROPERTYEX(d.name, 'IsFulltextEnabled') as IsFulltextEnabled,
                DATABASEPROPERTYEX(d.name, 'IsInStandBy') as IsInStandBy,
                DATABASEPROPERTYEX(d.name, 'IsLocalCursorsDefault') as IsLocalCursorsDefault,
                DATABASEPROPERTYEX(d.name, 'IsMemoryOptimizedElevateToSnapshotEnabled') as IsMemoryOptimizedElevateToSnapshotEnabled,
                DATABASEPROPERTYEX(d.name, 'IsMergePublished') as IsMergePublished,
                DATABASEPROPERTYEX(d.name, 'IsNullConcat') as IsNullConcat,
                DATABASEPROPERTYEX(d.name, 'IsNumericRoundAbortEnabled') as IsNumericRoundAbortEnabled,
                DATABASEPROPERTYEX(d.name, 'IsParameterizationForced') as IsParameterizationForced,
                DATABASEPROPERTYEX(d.name, 'IsQuotedIdentifiersEnabled') as IsQuotedIdentifiersEnabled,
                DATABASEPROPERTYEX(d.name, 'IsPublished') as IsPublished,
                DATABASEPROPERTYEX(d.name, 'IsRecursiveTriggersEnabled') as IsRecursiveTriggersEnabled,
                DATABASEPROPERTYEX(d.name, 'IsSubscribed') as IsSubscribed,
                DATABASEPROPERTYEX(d.name, 'IsSyncWithBackup') as IsSyncWithBackup,
                DATABASEPROPERTYEX(d.name, 'IsTornPageDetectionEnabled') as IsTornPageDetectionEnabled,
                DATABASEPROPERTYEX(d.name, 'IsVerifiedClone') as IsVerifiedClone,
                DATABASEPROPERTYEX(d.name, 'IsXTPSupported') as IsXTPSupported,
                DATABASEPROPERTYEX(d.name, 'LastGoodCheckDbTime') as LastGoodCheckDbTime,
                DATABASEPROPERTYEX(d.name, 'LCID') as LCID,
                DATABASEPROPERTYEX(d.name, 'MaxSizeInBytes') as MaxSizeInBytes,
                DATABASEPROPERTYEX(d.name, 'Recovery') as Recovery,
                DATABASEPROPERTYEX(d.name, 'ServiceObjective') as ServiceObjective,
                DATABASEPROPERTYEX(d.name, 'ServiceObjectiveId') as ServiceObjectiveId,
                DATABASEPROPERTYEX(d.name, 'SQLSortOrder') as SQLSortOrder,
                DATABASEPROPERTYEX(d.name, 'Status') as Status,
                DATABASEPROPERTYEX(d.name, 'Updateability') as Updateability,
                DATABASEPROPERTYEX(d.name, 'UserAccess') as UserAccess,
                DATABASEPROPERTYEX(d.name, 'Version') as Version,
                d.is_trustworthy_on,
                d.is_db_chaining_on,
                suser_sname(d.owner_sid) as Owner
              for xml path('Properties'), type
            ),
            (
                select 
                    file_id,
                    type,
                    type_desc,
                    data_space_id,
                    name,
                    physical_name,
                    state,
                    state_desc,
                    -- size, -- too volatile
                    max_size,
                    growth,
                    is_media_read_only,
                    is_read_only,
                    is_percent_growth
                from 
                    sys.master_files mf
                where mf.database_id = d.database_id
                order by mf.name
                for xml path ('File'), type
            ) as Files
        for xml path('DatabaseInfo'), type
    ) as DatabaseInfo
from sys.databases d
"""
        );
    }
    
    protected List<DatabaseInfo> getDatabases(Connection conn, String sql){
        Statement statement = null;
        ResultSet catalogs = null;
        try {
            statement = conn.createStatement();
            catalogs  = statement.executeQuery(sql);

            List<DatabaseInfo> result = new ArrayList<DatabaseInfo>();
            while (catalogs.next()) {
                def state = catalogs.getString("State")
                def readable = state.toUpperCase().matches("ONLINE")
                def dbInfo = new DatabaseInfo(catalogs.getString("name"),state,readable)
                dbInfo.setCustomData("State", state)
                dbInfo.setCustomData("Compatibility Level", catalogs.getString("Compatibility Level"))
                dbInfo.setCustomData("Create Date", catalogs.getString("Create Date"))
                dbInfo.setCustomData("Readonly", catalogs.getString("Readonly"))
                dbInfo.setCustomData("Recovery Model", catalogs.getString("Recovery Model"))
                dbInfo.setCustomData("DatabaseDefinition", catalogs.getString("DatabaseInfo"));

                result.add(dbInfo);
            }
            return result;
        } catch (ApiException e) {
            throw e;
        } catch (Exception e) {
            throw new ApiException("Cannot load databases:"+e.getLocalizedMessage(), e);
        } finally {
            closeQuietly(catalogs);
            closeQuietly(statement);
        }
    }
    
    public static String getScheduleSummary(GPathResult s){
        def type = s.freq_type.text() as int
        def interval  = s.freq_interval.text() as int
        def subdayType = s.freq_subday_type.text() as int
        def subdayInterval = s.freq_subday_interval.text() as int
        def relativeInterval = s.freq_relative_interval.text() as int
        def reccurence = s.freq_recurrence_factor.text() as int
        def startDate = s.active_start_date.text()
        def startTime = String.format("%06d", s.active_start_time.text() as int)
        def endDate = s.active_end_date.text()
        def endTime = String.format("%06d", s.active_end_time.text() as int)
        //System.out.println("EndDate="+endDate)
        def convertDateTime = { d, t ->
           def startDateTime = Date.parse('yyyyMMddHHmmss',d+t)
           return "${startDateTime.format('MM/dd/yyyy')} at ${startDateTime.format('hh:mm:ss a')}"
        }

        def convertTime = { t ->
           def startDateTime = Date.parse('HHmmss',t)
           return "${startDateTime.format('hh:mm:ss a')}"
        }
        def convertDate = { d ->
           def startDateTime = Date.parse('yyyyMMdd',d)
           return "${startDateTime.format('MM/dd/yyyy')}"
        }
        
        def r = ""
        if (type==1) {
            r += "Occurs once on "+convertDateTime(startDate,startTime)
        }
        if (type==4) {
            if (interval==1) {
                r += "Occurs every day"
            } else {
                r += "Occurs every ${interval} day(s)"
            }
        }
        if (type==8) {
            if (reccurence==1) {
                r += "Occurs every week on "
            } else {
                r += "Occurs every ${reccurence} week(s) on "
            }
            def days = []
            if ((interval&1)==1)   days << "Sunday"
            if ((interval&2)==2)   days << "Monday"
            if ((interval&4)==4)   days << "Tuesday"
            if ((interval&8)==8)   days << "Wednesday"
            if ((interval&16)==16) days << "Thursday"
            if ((interval&32)==32) days << "Friday"
            if ((interval&64)==64) days << "Saturday"
            r+= days.join(", ")
        }

        if (type==16) {
            if (reccurence==1) {
                r += "Occurs every month on "
            } else {
                r += "Occurs every ${reccurence} months(s) on "
            }
            r+=" day ${interval} of that month "
        }

        if (type==32) {
            r += "Occurs every "
            switch (relativeInterval) {
                case 1:
                    r+="first "
                    break
                case 2:
                    r+="second "
                    break
                case 4:
                    r+="third "
                    break
                case 8:
                    r+="fourth "
                    break
                case 16:
                    r+="last "
                    break
            }
            if (interval==1)  r+= "Sunday"
            if (interval==2)  r+= "Monday"
            if (interval==3)  r+= "Tuesday"
            if (interval==4)  r+= "Wednesday"
            if (interval==5)  r+= "Thursday"
            if (interval==6)  r+= "Friday"
            if (interval==7)  r+= "Saturday"
            if (interval==8)  r+= "day"
            if (interval==9)  r+= "weekday"
            if (interval==10) r+= "weekend day"
            r+=" of every ${reccurence} month(s)"
        }
        
        if (type>=4 && type<=32) {
            switch (subdayType) {
                case 1:
                    r+=" at "+ convertTime(startTime)
                    break
                case 2:
                    r+=" every ${subdayInterval} seconds(s) between ${convertTime(startTime)} and ${convertTime(endTime)}"
                    break
                case 4:
                    r+=" every ${subdayInterval} minute(s) between ${convertTime(startTime)} and ${convertTime(endTime)}"
                    break
                case 8:
                    r+=" every ${subdayInterval} hour(s) between ${convertTime(startTime)} and ${convertTime(endTime)}"
                    break
            }
        }
        
        if (type>=4 && type<=32) {
             r+=". Schedule will be used "
             if (endDate.equals("99991231")) {
                 r+=" starting on ${convertDate(startDate)}"
             } else {
                 r+=" between ${convertDate(startDate)} and ${convertDate(endDate)}"
             }
        }
      
        if (type==64) {
            r += "Start automatically when SQL Server Agent starts"
        }
        if (type==128) {
            r += "Start whenever the CPUs become idle"
        }
        return r
    }
}