import com.branegy.dbmaster.model.Function
import com.branegy.dbmaster.model.RevEngineeringOptions

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import com.branegy.dbmaster.connection.*
import com.branegy.dbmaster.model.*
import com.branegy.dbmaster.model.ForeignKey.ReferentialAction;
import com.branegy.dbmaster.model.ForeignKey.ColumnMapping;
import com.branegy.dbmaster.model.Index.IndexColumn;
import com.branegy.dbmaster.model.Index.IndexType;
import com.branegy.dbmaster.model.Parameter.ParamType;
import com.branegy.dbmaster.util.NameMap;
import com.branegy.service.core.exception.*
import com.branegy.dbmaster.connection.AbstractJdbcDialect.PreparedSql
import java.util.AbstractMap.SimpleEntry

/*
 * 
 * SQL 2000: User has access to all definitions of the database 
 * SQL 2005: user "GRANT VIEW DEFINITION TO [<user>]" for database
 */

// TODO (Slava) function is not implemented
public class SQLServer2000Dialect extends SQLServerDialect {
    final Set<String> excludeProcedureSet = [
        "dbo.sp_alterdiagram",
        "dbo.sp_creatediagram",
        "dbo.sp_dropdiagram",
        "dbo.sp_helpdiagramdefinition",
        "dbo.sp_helpdiagrams",
        "dbo.sp_renamediagram",
        "dbo.sp_upgraddiagrams"] as Set;
    final Set<String> excludeFunctionSet = ["dbo.fn_diagramobjects"] as Set;
    final Set<String> excludeTableSet = ["dbo.sysdiagrams"] as Set;

    public SQLServer2000Dialect(JdbcConnector cp,Connection connection, String dialectName, String dialectVersion)
            throws SQLException {
        super(cp,connection, dialectName, dialectVersion);
        typesWithPrecesion = new HashSet<String>();
        typesWithScale = new HashSet ( [ "NUMERIC", "DECIMAL" ] );
        typesWithSize = new HashSet   ([ "BINARY", "CHAR", "DATETIME2", "DATETIMEOFFSET", 
                                         "DECIMAL", "NCHAR", "NUMERIC", "NVARCHAR", "TIME", "VARBINARY", "VARCHAR" ]) 
    }

    @Override
    protected NameMap<Table> getTables(Connection conn, RevEngineeringOptions options) throws Exception {
        String sql =
                """ select s.name as schemaName,  
                           t.name as tableName,
                           null as ${Table.DESCRIPTION},
                           -- cast(p.value as varchar(2000)) as ${Table.DESCRIPTION},
                           t.crdate, 
                           modify_date = null, 
                           sysindexes.rows as ${Table.ROWS}              
                    from sysobjects as t 
                    inner join sysusers s ON t.uid = s.uid
                    left join  ::fn_listextendedproperty (NULL,'user','dbo','table','t.name',null,DEFAULT) as p 
                               on p.objname=t.name and p.name='MS_Description'       
                    left join sysindexes ON t.id = sysindexes.id AND sysindexes.indid < 2
                    where t.xtype='U' AND""";
        return this.getTables(conn, options, sql,"order by schemaName,tableName","s.name", "t.name",
            new SimpleEntry<>(Table.DESCRIPTION, String.class),
            new SimpleEntry<>(Table.ROWS, Integer.class),
        );
    }

    @Override
    protected void populateTableColumns(Connection conn, RevEngineeringOptions options, 
                                        NameMap<Table> tables) throws SQLException {

        String sql =
              """select sch.name as schema_name,tab.name as table_name,col.name as column_name,  
            systyp.name as type_name, typ.is_user_defined, typ.name as user_type, 
            col.max_length, col.precision, col.scale, col.collation_name, col.is_nullable as nullable, 
            col.is_identity,col.is_computed,cc.definition as computedExpression, col.is_rowguidcol,
            def.definition as defaultValue, cast(ep.value as varchar(7500)) as comment 
            from sys.columns as col  
            inner join sys.tables 
             as tab on col.object_id=tab.object_id  
            inner join sys.schemas as sch on tab.schema_id = sch.schema_id  
            inner join sys.types as typ on col.user_type_id= typ.user_type_id  
            inner join sys.types as systyp on col.system_type_id  = systyp.system_type_id  
            and systyp.is_user_defined=0 and systyp.user_type_id = systyp.system_type_id  
            left join sys.default_constraints as def on col.default_object_id = def.object_id 
            left join sys.extended_properties as ep on ep.major_id=tab.object_id 
            and ep.minor_id=col.column_id 
            and ep.class='1' and ep.class_desc='OBJECT_OR_COLUMN' and ep.name='MS_Description' 
            left join sys.computed_columns as cc 
             on tab.object_id=cc.object_id and col.column_id = cc.column_id
            where""";

        this.populateTableColumns(conn,options,tables,sql,"order by schema_name, table_name, col.column_id",
            "sch.name", "tab.name");
        
        /*
        String extra = "";
        //            collation_name
        if (columnsRS.getInt("is_identity") == 1) {
            extra += " Identity";
        }
        if (columnsRS.getInt("is_computed") == 1) {
            extra += " Computed:" + columnsRS.getString("computedExpression");
        }
        if (columnsRS.getInt("is_rowguidcol") == 1) {
            extra += " RowGUID";
        }
        if (columnsRS.getInt("is_user_defined")==1) {
            extra += " User type:" + columnsRS.getString("user_type");
        }
        columnInfo.setExtraDefinition(extra);*/
        
    }

    @Override
    protected NameMap<View> getViews(Connection conn, RevEngineeringOptions options) throws Exception {
        String sql =
                """    select s.name as schemaName, v.name as viewName,  
                            cast(p.value as varchar(2000)) as viewDescription,  
                            c.text as viewText,c.colid, v.create_date, v.modify_date   
                       from sys.views as v  inner join sys.schemas as s   
                            on v.schema_id=s.schema_id inner join sys.syscomments as c   
                            on v.object_id=c.id left join sys.extended_properties as p   
                            on v.object_id = p.major_id and p.name='MS_Description'   
                            and p.minor_id=0
                       where v.type='V' AND""";

        return this.getViews(conn, options,sql,"order by s.name, v.name, colid","s.name", "v.name");
    }
    
    @Override
    protected NameMap<Procedure> getProcedures(Connection conn, RevEngineeringOptions options) throws Exception {
        String sql =
                """  select s.name as schemaName,  pr.name as procedureName, 
                cast(p.value as varchar(2000)) as procedureDescription,pr.create_date, pr.modify_date, 
                c.colid,c.text as procedureText from sys.procedures as pr 
                inner join sys.schemas as s on pr.schema_id=s.schema_id 
                left join sys.extended_properties as p on pr.object_id = p.major_id 
                and p.name='MS_Description' and p.minor_id=0 
                inner join sys.syscomments as c on pr.object_id=c.id
                where""";

        return this.getProcedures(conn, options, sql, "order by s.name, pr.name,c.colid", "s.name", "pr.name",
             "N'P',N'PC',N'X',N'RF'");
    }

    protected void populateParameters(Connection conn, RevEngineeringOptions options,
                                      NameMap dbObjects, String typeFilter) {

        String sql =
                """   select sch.name as schema_name,tab.name as procedure_name,col.name as column_name,
                      systyp.name as type_name,
                   col.max_length, col.precision, col.scale, col.is_xml_document, 
                   col.is_readonly, col.is_output,
                   cast(col.default_value as varchar(7500)) as defaultValue,
                   cast(ep.value as varchar(7500)) as comment
                   from sys.parameters as col 
                   inner join sys.procedures as tab on col.object_id=tab.object_id 
                   inner join sys.schemas as sch on tab.schema_id = sch.schema_id 
                   inner join sys.types as typ on col.user_type_id= typ.user_type_id 
                   inner join sys.types as systyp on col.system_type_id  = systyp.system_type_id 
                   and systyp.is_user_defined=0 and systyp.user_type_id = systyp.system_type_id 
                   left join sys.extended_properties as ep on ep.major_id=tab.object_id 
                           and ep.minor_id=col.parameter_id
                   and ep.class='1' and ep.class_desc='OBJECT_OR_COLUMN' and ep.name='MS_Description'
                   where""";

       this.populateParameters(conn,options,dbObjects,typeFilter, sql, null, "sch.name","col.name");
    }

    /**
     * @objType can be U: USER TABLE, V: View, TF: Table valued functions
     * Requires View server state permission
     */
     protected void populateIndexes(Connection conn, RevEngineeringOptions options, NameMap<Table> tables)
            throws Exception {
                
        String objType = 'U';
        String sql =
              """SELECT s.name as schemaName,o.name as objectName,o.type as objectType, 
                      i.name as indexName,i.type indexType,i.is_unique,i.ignore_dup_key, i.is_primary_key, 
                      i.is_unique_constraint,i.is_disabled, i.fill_factor, 
                      CASE WHEN ps.usedpages > ps.pages  
                           THEN (ps.usedpages - ps.pages)  
                           ELSE 0 END * 8 indexSize,  
                      stat.avg_fragmentation_in_percent as fragmentation  
                FROM  sys.objects o   
                inner join sys.indexes i on i.object_id = o.object_id  
                inner join sys.schemas s on o.schema_id = s.schema_id  
                 left join  sys.dm_db_index_physical_stats(DB_ID(), null, null, NULL , 'LIMITED') as stat  
                           on i.object_id = stat.object_id  and i.index_id = stat.index_id  
                left JOIN (  
                     SELECT  object_id, index_id, sum (used_page_count) usedpages,  
                    SUM (  
                      CASE  
                         WHEN (index_id < 2)   
                         THEN (in_row_data_page_count + lob_used_page_count + row_overflow_used_page_count)  
                         ELSE lob_used_page_count + row_overflow_used_page_count  
                      END ) pages  FROM sys.dm_db_partition_stats 
                  GROUP BY object_id, index_id  
              ) ps ON i.index_id = ps.index_id and i.object_id = ps.object_id  
              where i.name IS NOT NULL AND i.is_hypothetical=0 and o.type='${objType}' AND""";

          this.populateIndexes(conn, options, tables, sql, null,"s.name","i.name");
    }
    
    @Override
    protected  NameMap<Function> getFunctions(Connection conn, RevEngineeringOptions options) {
        return new NameMap<Function>();
    }
    
            
    @Override
    protected void setTableColumnComments(Connection conn, String database, 
            NameMap<Table> tables) throws SQLException {
        // NOP
    }
    
    @Override
    protected void setViewColumnComments(Connection conn, String database, 
            NameMap<View> tables) throws SQLException {
        // NOP
    }
}
/*

SELECT * FROM databases

SELECT * FROM master.dbo.sysdatabases
sp_helpdb

SELECT * FROM master.dbo.sysdatabases

select * from INFORMATION_SCHEMA.SCHEMATA


select name as DBNAME,STATUS,
CASE WHEN (STATUS &         1) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [AUTOCLOSE_(ALTER_DATABASE)],
CASE WHEN (STATUS &         4) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [SELECT_INTO/BULKCOPY_(ALTER_DATABASE_USING_SET_RECOVERY)],
CASE WHEN (STATUS &         8) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [TRUNC._LOG_ON_CHKPT_(ALTER_DATABASE_USING_SET_RECOVERY)],
CASE WHEN (STATUS &        16) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [TORN_PAGE_DETECTION_(ALTER_DATABASE)],
CASE WHEN (STATUS &        32) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [LOADING],
CASE WHEN (STATUS &        64) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [PRE_RECOVERY],
CASE WHEN (STATUS &       128) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [RECOVERING],
CASE WHEN (STATUS &       256) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [NOT_RECOVERED],
CASE WHEN (STATUS &       512) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [OFFLINE_(ALTER_DATABASE)],
CASE WHEN (STATUS &      1024) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [READ_ONLY_(ALTER_DATABASE)],
CASE WHEN (STATUS &      2048) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [DBO_USE_ONLY_(ALTER_DATABASE_USING_SET_RESTRICTED_USER)],
CASE WHEN (STATUS &      4096) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [SINGLE_USER_(ALTER_DATABASE)],
CASE WHEN (STATUS &     32768) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [EMERGENCY_MODE],
CASE WHEN (STATUS &   4194304) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [AUTOSHRINK_(ALTER_DATABASE)],
CASE WHEN (STATUS &1073741824) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [CLEANLY_SHUTDOWN]
from master.dbo.sysdatabases

SELECT name as DBNAME,STATUS2,
CASE WHEN (STATUS2 &     16384) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [ANSI_null_default_(ALTER_DATABASE)],
CASE WHEN (STATUS2 &     65536) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [concat_null_yields_null_(ALTER_DATABASE)],
CASE WHEN (STATUS2 &    131072) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [recursive_triggers_(ALTER_DATABASE)],
CASE WHEN (STATUS2 &   1048576) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [default_to_local_cursor_(ALTER_DATABASE)],
CASE WHEN (STATUS2 &   8388608) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [quoted_identifier_(ALTER_DATABASE)],
CASE WHEN (STATUS2 &  33554432) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [cursor_close_on_commit_(ALTER_DATABASE)],
CASE WHEN (STATUS2 &  67108864) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [ANSI_nulls_(ALTER_DATABASE)],
CASE WHEN (STATUS2 & 268435456) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [ANSI_warnings_(ALTER_DATABASE)],
CASE WHEN (STATUS2 & 536870912) = 0 THEN 'FALSE' ELSE 'TRUE' END AS [full_text_enabled_(set_by_using_sp_fulltext_database)]
from master.dbo.sysdatabases

*/