package io.dbmaster.sync;

import java.util.ArrayList
import java.util.Iterator
import java.util.List
import java.util.Map.Entry

import org.slf4j.Logger
import org.slf4j.LoggerFactory

import com.branegy.dbmaster.sync.api.*
import com.branegy.dbmaster.sync.impl.*
import com.branegy.dbmaster.sync.api.SyncPair.ChangeType
import com.branegy.dbmaster.database.api.ModelService
import com.branegy.scripting.DbMaster
import com.branegy.service.core.QueryRequest
import com.branegy.dbmaster.sync.api.SyncAttributePair.SyncAttributeComparator;
import com.branegy.dbmaster.sync.api.SyncSession.SearchTarget
import com.branegy.dbmaster.sync.api.SyncService
import com.branegy.dbmaster.model.*
import com.branegy.email.EmailSender;


ModelService modelService = dbm.getService(ModelService.class);
List<Model> models = modelService.getModelList(null, null);

Map<String,String> attachments = [:] as LinkedHashMap<>;
for (Model model:models) {
    for (ModelDataSource ds:model.getDataSources()) {
        logger.info("Synchronize {} {}", model.getName(), ds.getName())
        SyncSession syncSession = modelService.synchronizeModel(ds.getId());
        String html = modelService.generateSyncSessionPreviewHtml(syncSession, true);
        logger.debug("done")
        if (syncSession.getSyncResult().getChangeType() != ChangeType.EQUALS) {
            syncSession.applyChanges();
            attachments.put(model.getName()+"."+ds.getName(), html);
            logger.info("Changes are found")
        }
    }
}

if (!attachments.isEmpty()) {
    EmailSender sender = dbm.getService(EmailSender.class);
    String msg = 
"""
Changes are found in ${attachments.keySet()}
   
---------------------------------------
This message was automatically generated by DBMaster plugin
""";
    sender.createMessage(p_email, "Model Sync Changes", msg, true);
    attachments.forEach({k,v ->
        sender.addAttachment(k+".html",v);
    });
    sender.sendMessage();
} else {
    println "No changes are found";
}

